/*! \file
\verbatim

Copyright (c) 2004, Sylvain Paris and Francois Sillion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    * Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    * Neither the name of ARTIS, GRAVIR-IMAG nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

\endverbatim
*
*  This file defines some edge-detecting functions.
*/

#ifndef __EDGE_DETECTOR__
#define __EDGE_DETECTOR__


#include "base_2D.h"

/*!
  \file
  Contains edge-detector functions. Derivative of Gaussian is inside
  gaussian.h .
*/

namespace Function_2D{


  /*
    #################
    # Canny Deriche #
    #################
  */

  //! \f$ -x.exp(-(|x|+\frac{1}{2}y^2)) \f$ Not normalized.
  class Canny_Deriche:public Filter_interface_2D{

  public:
    inline Canny_Deriche(const real_type center_x = 0,
			 const real_type center_y = 0,
			 const real_type sigma_x = 1,
			 const real_type sigma_y = 1,
			 const real_type orientation_angle = 0);

  protected:
    virtual inline real_type base_value_proxy(const real_type x,
					      const real_type y) const;

  };



  /*
    ###############
    # Shen Castan #
    ###############
  */

  //! \f$ -\frac{x}{|x|}exp(-(|x|+\frac{1}{2}y^2)) \f$
  class Shen_Castan:public Filter_interface_2D{

  public:
    inline Shen_Castan(const real_type center_x = 0,
		       const real_type center_y = 0,
		       const real_type sigma_x = 1,
		       const real_type sigma_y = 1,
		       const real_type orientation_angle = 0);

  protected:
    virtual inline real_type base_value_proxy(const real_type x,
					      const real_type y) const;

  };

  
/*
  ###########################################################
  ###########################################################
  ###########################################################
  ##############                               ##############
  ##############  I M P L E M E N T A T I O N  ##############
  ##############                               ##############
  ###########################################################
  ###########################################################
  ###########################################################
*/


/*
  #################
  # Canny Deriche #
  #################
*/
  Canny_Deriche::
  Canny_Deriche(const real_type center_x,
		const real_type center_y,
		const real_type sigma_x,
		const real_type sigma_y,
		const real_type orientation_angle):
    Filter_interface_2D(center_x,center_y,
			sigma_x,sigma_y,
			1,1,
			orientation_angle,
			1){}

  Canny_Deriche::real_type
  Canny_Deriche::base_value_proxy(const real_type x,
				  const real_type y) const{
    
    return -x*exp(-fabs(x)-0.5*y*y);
  }


  /*
    ###############
    # Shen Castan #
    ###############
  */
  
  Shen_Castan::Shen_Castan(const real_type center_x,
			   const real_type center_y,
			   const real_type sigma_x,
			   const real_type sigma_y,
			   const real_type orientation_angle):
    Filter_interface_2D(center_x,center_y,
			sigma_x,sigma_y,
			1,1,
			orientation_angle,
			1){}
  
  Shen_Castan::real_type
  Shen_Castan::base_value_proxy(const real_type x,
				const real_type y) const{

    const real_type s = (x==0) ? 0 : ((x>0) ? 1 : -1);
    
    return -s*exp(-fabs(x)-0.5*y*y);
  }
}

#endif

