/*! \file
\verbatim

Copyright (c) 2004, Sylvain Paris and Francois Sillion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    * Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    * Neither the name of ARTIS, GRAVIR-IMAG nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

\endverbatim
*
*  This file defines the Gabor functions.
*/

#ifndef __GABOR__
#define __GABOR__

#include "base_2D.h"
#include "gaussian_2D.h"

namespace Function_2D{

//! Gabor filter (normalized gaussian modulated by a sine: \f$ cos(\frac{2\pi x}{\lambda}+\phi)exp(-r^2) \f$).
  class Gabor:public Filter_interface_2D{

  public:
    Gabor::Gabor(const real_type center_x = 0,
		 const real_type center_y = 0,
		 const real_type sigma_x = 1,
		 const real_type sigma_y = 1,
		 const real_type orientation_angle = 0,
		 const real_type w_phase = 0,
		 const real_type w_length = 1);

    inline void set_phase(const real_type p);
    inline void set_wave_length(const real_type w);

  protected:

    //! Phase of the sine.
    real_type phase;

    //! Wave length of the sine divided by sigma_x (i.e. the wave length
    //! is scaled consistently according to sigma_x).
    real_type wave_length;

    
    virtual inline real_type base_value_proxy(const real_type x,
					      const real_type y) const;


  };



  /*
    ###########################################################
    ###########################################################
    ###########################################################
    ##############                               ##############
    ##############  I M P L E M E N T A T I O N  ##############
    ##############                               ##############
    ###########################################################
    ###########################################################
    ###########################################################
  */

  Gabor::Gabor(const real_type center_x,
	       const real_type center_y,
	       const real_type sigma_x,
	       const real_type sigma_y,
	       const real_type orientation_angle,
	       const real_type w_phase,
	       const real_type w_length):
    Filter_interface_2D(center_x,center_y,
			sigma_x,sigma_y,
			M_SQRT1_2,M_SQRT1_2,
			orientation_angle,
			0.5/M_PI),
    phase(w_phase),
    wave_length(w_length){}
  
  void Gabor::set_phase(const real_type p){
    phase = p;
  }

  void Gabor::set_wave_length(const real_type w){
    wave_length = w;
  }

  Gabor::real_type
  Gabor::base_value_proxy(const real_type x,
			  const real_type y) const{

    return cos(2*M_PI*x/wave_length + phase) * Gaussian::base_value(x,y);
  }
  
}

#endif
