/*! \file
\verbatim

Copyright (c) 2004, Sylvain Paris and Francois Sillion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    * Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    * Neither the name of ARTIS, GRAVIR-IMAG nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

\endverbatim
*
*  This file defines some Gaussian-based functions.
*/

#ifndef __GAUSSIAN_2D__
#define __GAUSSIAN_2D__

#include <math.h>

#include "base_2D.h"

/*!
  \file
  Contains Gaussian-related functions.

*/

namespace Function_2D{

/*
  ############
  # Gaussian #
  ############
*/

  
//! Very basic gaussian function \f$ exp(-r^2) \f$.
  class Gaussian:public Base_2D{

  public:
    inline Gaussian::Gaussian(const real_type in_offset_x = 0,
			      const real_type in_offset_y = 0,
			      const real_type in_scale_x = 1,
			      const real_type in_scale_y = 1,
			      const real_type in_rotation_angle = 0,
			      const real_type out_offset = 0,
			      const real_type out_scale = 1);

    static inline real_type base_value(const real_type x,
				       const real_type y);

  protected:
    virtual inline real_type base_value_proxy(const real_type x,
					      const real_type y) const;


  
  };




  
/*
  #######################
  # Normalized gaussian #
  #######################
*/


//! Gaussian function normalized to 1.
  class Normalized_gaussian:public Filter_interface_2D{

  public:
    inline Normalized_gaussian(const real_type center_x = 0,
			       const real_type center_y = 0,
			       const real_type sigma_x = 1,
			       const real_type sigma_y = 1,
			       const real_type orientation_angle = 0);

  protected:
    virtual inline real_type base_value_proxy(const real_type x,
					      const real_type y) const;
  };



/*
  ###########################
  # FFT normalized gaussian #
  ###########################
*/


//! FFT of a gaussian function normalized to 1.
  class FFT_normalized_gaussian:public Filter_interface_2D{

  public:
    inline FFT_normalized_gaussian(const real_type center_x = 0,
 				   const real_type center_y = 0,
				   const real_type sigma_x = 1,
				   const real_type sigma_y = 1,
				   const real_type orientation_angle = 0);

  protected:
    virtual inline real_type base_value_proxy(const real_type x,
					      const real_type y) const;
  };



  
/*
  #########################
  # Gaussian x-derivative #
  #########################
*/


//! \f$ -2x.exp(-r^2) \f$.
  class Gaussian_x_derivative:public Base_2D{

  public:
    inline Gaussian_x_derivative::
    Gaussian_x_derivative(const real_type in_offset_x = 0,
			  const real_type in_offset_y = 0,
			  const real_type in_scale_x = 1,
			  const real_type in_scale_y = 1,
			  const real_type in_rotation_angle = 0,
			  const real_type out_offset = 0,
			  const real_type out_scale = 1);

    static inline real_type base_value(const real_type x,
				       const real_type y);

  protected:
    virtual inline real_type base_value_proxy(const real_type x,
					      const real_type y) const;


  
  };


  
/*
  ####################################
  # Normalized gaussian x-derivative #
  ####################################
*/


//! X-derivative of a gaussian function normalized to 1.
  class Normalized_gaussian_x_derivative:public Filter_interface_2D{

  public:
    inline Normalized_gaussian_x_derivative(const real_type center_x = 0,
					    const real_type center_y = 0,
					    const real_type sigma_x = 1,
					    const real_type sigma_y = 1,
					    const real_type orientation_angle = 0);

  protected:
    virtual inline real_type base_value_proxy(const real_type x,
					      const real_type y) const;
  };





  
/*
  #########################
  # Gaussian y-derivative #
  #########################
*/


//! \f$ -2y.exp(-r^2) \f$.
  class Gaussian_y_derivative:public Base_2D{

  public:
    inline Gaussian_y_derivative::
    Gaussian_y_derivative(const real_type in_offset_x = 0,
			  const real_type in_offset_y = 0,
			  const real_type in_scale_x = 1,
			  const real_type in_scale_y = 1,
			  const real_type in_rotation_angle = 0,
			  const real_type out_offset = 0,
			  const real_type out_scale = 1);

    static inline real_type base_value(const real_type x,
				       const real_type y);

  protected:
    virtual inline real_type base_value_proxy(const real_type x,
					      const real_type y) const;


  
  };


  
/*
  ####################################
  # Normalized gaussian y-derivative #
  ####################################
*/


//! X-derivative of a gaussian function normalized to 1.
  class Normalized_gaussian_y_derivative:public Filter_interface_2D{

  public:
    inline Normalized_gaussian_y_derivative(const real_type center_x = 0,
					    const real_type center_y = 0,
					    const real_type sigma_x = 1,
					    const real_type sigma_y = 1,
					    const real_type orientation_angle=0);

  protected:
    virtual inline real_type base_value_proxy(const real_type x,
					      const real_type y) const;
  };



  
/*
  #############################
  # Gaussian x-2nd-derivative #
  #############################
*/


//! \f$ (4x^2-2).exp(-r^2) \f$.
  class Gaussian_x_2nd_derivative:public Base_2D{

  public:
    inline Gaussian_x_2nd_derivative::
    Gaussian_x_2nd_derivative(const real_type in_offset_x = 0,
			      const real_type in_offset_y = 0,
			      const real_type in_scale_x = 1,
			      const real_type in_scale_y = 1,
			      const real_type in_rotation_angle = 0,
			      const real_type out_offset = 0,
			      const real_type out_scale = 1);
    
    static inline real_type base_value(const real_type x,
				       const real_type y);

  protected:
    virtual inline real_type base_value_proxy(const real_type x,
					      const real_type y) const;


  
  };


  
/*
  ########################################
  # Normalized gaussian x-2nd_derivative #
  ########################################
*/


//! Second X-derivative of a gaussian function normalized to 1.
  class Normalized_gaussian_x_2nd_derivative:public Filter_interface_2D{

  public:
    inline Normalized_gaussian_x_2nd_derivative(const real_type center_x = 0,
						const real_type center_y = 0,
						const real_type sigma_x = 1,
						const real_type sigma_y = 1,
						const real_type orientation_angle = 0);
    
  protected:
    virtual inline real_type base_value_proxy(const real_type x,
					      const real_type y) const;
  };







  
  /*!
    #################
    # Gaussian step #
    #################
  */

  

//! \f$ \frac{x}{|x|}exp(-r^2) \f$.
  class Gaussian_step:public Base_2D{

  public:
    inline Gaussian_step::Gaussian_step(const real_type in_offset_x = 0,
					const real_type in_offset_y = 0,
					const real_type in_scale_x = 1,
					const real_type in_scale_y = 1,
					const real_type in_rotation_angle = 0,
					const real_type out_offset = 0,
					const real_type out_scale = 1);

    static inline real_type base_value(const real_type x,
				       const real_type y);

  protected:
    virtual inline real_type base_value_proxy(const real_type x,
					      const real_type y) const;


  
  };




  
/*
  ############################
  # Normalized gaussian step #
  ############################
*/


//! Gaussian function normalized to 1 multiplied by \f$ \frac{x}{|x|} \f$.
  class Normalized_gaussian_step:public Filter_interface_2D{

  public:
    inline Normalized_gaussian_step(const real_type center_x = 0,
				    const real_type center_y = 0,
				    const real_type sigma_x = 1,
				    const real_type sigma_y = 1,
				    const real_type orientation_angle = 0);

  protected:
    virtual inline real_type base_value_proxy(const real_type x,
					      const real_type y) const;
  };


  

/*
  ###########################################################
  ###########################################################
  ###########################################################
  ##############                               ##############
  ##############  I M P L E M E N T A T I O N  ##############
  ##############                               ##############
  ###########################################################
  ###########################################################
  ###########################################################
*/




/*
  ############
  # Gaussian #
  ############
*/
  
  Gaussian::Gaussian(const real_type in_offset_x,
		     const real_type in_offset_y,
		     const real_type in_scale_x,
		     const real_type in_scale_y,
		     const real_type in_rotation_angle,
		     const real_type out_offset,
		     const real_type out_scale):
    Base_2D(in_offset_x,in_offset_y,
	    in_scale_x,in_scale_y,
	    in_rotation_angle,
	    out_offset,out_scale){}

  Gaussian::real_type Gaussian::base_value_proxy(const real_type x,
						 const real_type y) const{

    return Gaussian::base_value(x,y);
  }

  Gaussian::real_type Gaussian::base_value(const real_type x,
					   const real_type y){
    return exp(-(x*x + y*y));
  }



  
/*
  #######################
  # Normalized gaussian #
  #######################
*/

  Normalized_gaussian::
  Normalized_gaussian(const real_type center_x,
		      const real_type center_y,
		      const real_type sigma_x,
		      const real_type sigma_y,
		      const real_type orientation_angle):
    Filter_interface_2D(center_x,center_y,
			sigma_x,sigma_y,
			M_SQRT1_2,M_SQRT1_2,
			orientation_angle,
			0.5/M_PI){}

  Normalized_gaussian::real_type
  Normalized_gaussian::base_value_proxy(const real_type x,
					const real_type y) const{

    return Gaussian::base_value(x,y);
  }



/*
  ###########################
  # FFT normalized gaussian #
  ###########################
*/

  FFT_normalized_gaussian::
  FFT_normalized_gaussian(const real_type center_x,
			  const real_type center_y,
			  const real_type sigma_x,
			  const real_type sigma_y,
			  const real_type orientation_angle):
    Filter_interface_2D(center_x,center_y,
			sigma_x,sigma_y,
			M_SQRT2*M_PI,M_SQRT2*M_PI,
			orientation_angle,
			1){
    output_scale = 1;
  }

  FFT_normalized_gaussian::real_type
  FFT_normalized_gaussian::base_value_proxy(const real_type x,
					    const real_type y) const{
    
    FFT_normalized_gaussian* proxy = const_cast<FFT_normalized_gaussian*>(this);
    proxy->output_scale = 1;
    return Gaussian::base_value(x,y);
  }



  /*
    #########################
    # Gaussian x-derivative #
    #########################
  */

  Gaussian_x_derivative::Gaussian_x_derivative(const real_type in_offset_x,
					       const real_type in_offset_y,
					       const real_type in_scale_x,
					       const real_type in_scale_y,
					       const real_type in_rotation_angle,
					       const real_type out_offset,
					       const real_type out_scale):
    Base_2D(in_offset_x,in_offset_y,
	    in_scale_x,in_scale_y,
	    in_rotation_angle,
	    out_offset,out_scale){}

  Gaussian_x_derivative::real_type
  Gaussian_x_derivative::base_value_proxy(const real_type x,
					  const real_type y) const{
    
    return Gaussian_x_derivative::base_value(x,y);
  }

  Gaussian_x_derivative::real_type
  Gaussian_x_derivative::base_value(const real_type x,
				    const real_type y){
    
    return -2*x*exp(-(x*x + y*y));
  }


  

  /*
    ####################################
    # Normalized gaussian x-derivative #
    ####################################
  */
  
  Normalized_gaussian_x_derivative::
  Normalized_gaussian_x_derivative(const real_type center_x,
				   const real_type center_y,
				   const real_type sigma_x,
				   const real_type sigma_y,
				   const real_type orientation_angle):
    Filter_interface_2D(center_x,center_y,
			sigma_x,sigma_y,
			M_SQRT1_2,M_SQRT1_2,
			orientation_angle,
			0.5/M_PI){}

  Normalized_gaussian_x_derivative::real_type
  Normalized_gaussian_x_derivative::base_value_proxy(const real_type x,
						     const real_type y) const{

    return Gaussian_x_derivative::base_value(x,y)*input_scale_x;
  }

  

  /*
    #########################
    # Gaussian y-derivative #
    #########################
  */

  Gaussian_y_derivative::Gaussian_y_derivative(const real_type in_offset_x,
					       const real_type in_offset_y,
					       const real_type in_scale_x,
					       const real_type in_scale_y,
					       const real_type in_rotation_angle,
					       const real_type out_offset,
					       const real_type out_scale):
    Base_2D(in_offset_x,in_offset_y,
	    in_scale_x,in_scale_y,
	    in_rotation_angle,
	    out_offset,out_scale){}

  Gaussian_y_derivative::real_type
  Gaussian_y_derivative::base_value_proxy(const real_type x,
					  const real_type y) const{
    
    return Gaussian_y_derivative::base_value(x,y);
  }

  Gaussian_y_derivative::real_type
  Gaussian_y_derivative::base_value(const real_type x,
				    const real_type y){
    
    return -2*y*exp(-(x*x + y*y));
  }


  

  /*
    ####################################
    # Normalized gaussian y-derivative #
    ####################################
  */
  
  Normalized_gaussian_y_derivative::
  Normalized_gaussian_y_derivative(const real_type center_x,
				   const real_type center_y,
				   const real_type sigma_x,
				   const real_type sigma_y,
				   const real_type orientation_angle):
    Filter_interface_2D(center_x,center_y,
			sigma_x,sigma_y,
			M_SQRT1_2,M_SQRT1_2,
			orientation_angle,
			0.5/M_PI){}

  Normalized_gaussian_y_derivative::real_type
  Normalized_gaussian_y_derivative::base_value_proxy(const real_type x,
						     const real_type y) const{
    
    return Gaussian_y_derivative::base_value(x,y)*input_scale_x;
  }



  /*
    #############################
    # Gaussian x-2nd_derivative #
    #############################
  */

  Gaussian_x_2nd_derivative::Gaussian_x_2nd_derivative(const real_type in_offset_x,
					       const real_type in_offset_y,
					       const real_type in_scale_x,
					       const real_type in_scale_y,
					       const real_type in_rotation_angle,
					       const real_type out_offset,
					       const real_type out_scale):
    Base_2D(in_offset_x,in_offset_y,
	    in_scale_x,in_scale_y,
	    in_rotation_angle,
	    out_offset,out_scale){}

  Gaussian_x_2nd_derivative::real_type
  Gaussian_x_2nd_derivative::base_value_proxy(const real_type x,
					  const real_type y) const{
    
    return Gaussian_x_2nd_derivative::base_value(x,y);
  }

  Gaussian_x_2nd_derivative::real_type
  Gaussian_x_2nd_derivative::base_value(const real_type x,
					const real_type y){
    
    return (4*x*x-2)*exp(-(x*x + y*y));
  }


  
  

  /*
    ########################################
    # Normalized gaussian x-2nd_derivative #
    ########################################
  */
  
  Normalized_gaussian_x_2nd_derivative::
  Normalized_gaussian_x_2nd_derivative(const real_type center_x,
				       const real_type center_y,
				       const real_type sigma_x,
				       const real_type sigma_y,
				       const real_type orientation_angle):
    Filter_interface_2D(center_x,center_y,
			sigma_x,sigma_y,
			M_SQRT1_2,M_SQRT1_2,
			orientation_angle,
			0.5/M_PI){}

  Normalized_gaussian_x_2nd_derivative::real_type
  Normalized_gaussian_x_2nd_derivative::base_value_proxy(const real_type x,
							 const real_type y) const{
    
    return Gaussian_x_2nd_derivative::base_value(x,y)*input_scale_x*input_scale_x;
  }

  
  

  /*
    #################
    # Gaussian step #
    #################
  */
  
  Gaussian_step::Gaussian_step(const real_type in_offset_x,
			       const real_type in_offset_y,
			       const real_type in_scale_x,
			       const real_type in_scale_y,
			       const real_type in_rotation_angle,
			       const real_type out_offset,
			       const real_type out_scale):
    Base_2D(in_offset_x,in_offset_y,
	    in_scale_x,in_scale_y,
	    in_rotation_angle,
	    out_offset,out_scale){}

  Gaussian_step::real_type
  Gaussian_step::base_value_proxy(const real_type x,
				  const real_type y) const{
    
    return Gaussian_step::base_value(x,y);
  }

  Gaussian_step::real_type Gaussian_step::base_value(const real_type x,
						     const real_type y){

    const real_type s = (x==0) ? 0 : ((x>0) ? 1 : -1);
    return s*exp(-(x*x + y*y));
  }



  
/*
  ############################
  # Normalized gaussian step #
  ############################
*/

  Normalized_gaussian_step::
  Normalized_gaussian_step(const real_type center_x,
			   const real_type center_y,
			   const real_type sigma_x,
			   const real_type sigma_y,
			   const real_type orientation_angle):
    Filter_interface_2D(center_x,center_y,
			sigma_x,sigma_y,
			M_SQRT1_2,M_SQRT1_2,
			orientation_angle,
			0.5/M_PI){}

  Normalized_gaussian_step::real_type
  Normalized_gaussian_step::base_value_proxy(const real_type x,
					     const real_type y) const{
    
    return Gaussian_step::base_value(x,y);
  }


}
  
#endif
